from django.db import models
from django.contrib import admin
from django.utils import timezone
from django.db.models import Model 
#from datetime import datetime

#to properly link User and Profile tables
from django.contrib.auth.models import User
from django.db.models.signals import post_save
from django.dispatch import receiver

#for a function in the cart model - May not need it at all yet.
from django.urls import reverse

#import choices list for model and forms and Variables
from .assets.choices import profile_status_choices, pedido_status_choices, pedido_frete_choices, pagamento_status_choices, billing_type_choices
from .assets.variables import Variables
from .assets.networkfunctions import add_days, normalizar

#For some reason, vscode does not like this import, but all works fine
from localflavor.br.models import BRStateField, BRCNPJField, BRPostalCodeField, BRCPFField

# Create your models here.

#Tabelas de dados especificos da Loja
class Loja(models.Model):
    cnpj = BRCNPJField(verbose_name="CNPJ") #coming from localflavor.br
    nome = models.CharField(max_length=150, verbose_name="Nome da loja")
    logo = models.ImageField(upload_to= 'images/', blank=True, null=True, verbose_name="Logo da loja") # this will go into the media default path
    imageminicio = models.ImageField(upload_to= 'images/', blank=True, null=True, verbose_name="Imagem da página inicial")
    imagemaluga = models.ImageField(upload_to= 'images/', blank=True, null=True, verbose_name="Imagem da página de aluguel")
    imagemfaq = models.ImageField(upload_to= 'images/', blank=True, null=True, verbose_name="Imagem da página de dúvidas")
    intro = models.TextField(blank=True, null=True, verbose_name="Introdução da loja")
    asaastoken = models.TextField(blank=True, null=True, verbose_name="Token da API do Asaas")
    asaaswebhook = models.TextField(blank=True, null=True, verbose_name="Token do webhook do Asaas")
    endereco = models.CharField(max_length=150,verbose_name="Endereço")
    complemento = models.CharField(max_length=50, verbose_name="Complemento")
    cep = BRPostalCodeField(verbose_name="CEP") #coming from localflavor.br
    bairro = models.CharField(max_length=30, verbose_name="Bairro")
    cidade = models.CharField(max_length=30, verbose_name="Cidade")
    estado = BRStateField(default="RJ", verbose_name="Estado") # coming from localflavor.br
    #pix = models.CharField(max_length=30, blank=True, null=True)
    #pixqrcode = models.ImageField(upload_to= 'images/', blank=True, null=True) # May not be needed as each transaction should get one
    email = models.EmailField(max_length=254, blank=True, null=True, verbose_name="Email")
    instagram = models.CharField(max_length=150, blank=True, null=True, verbose_name="Instagram")
    x = models.CharField(max_length=150, blank=True, null=True,verbose_name="X")
    facebook = models.CharField(max_length=150, blank=True, null=True,verbose_name="Facebook")
    tiktok = models.CharField(max_length=150, blank=True, null=True,verbose_name="TikTok")
    whatsapp = models.CharField(max_length=150, blank=True, null=True, verbose_name="Whatsapp")
    telegram = models.CharField(max_length=150, blank=True, null=True, verbose_name="Telegram")
    youtube = models.CharField(max_length=150, blank=True, null=True,verbose_name="Youtube")
    telefone = models.CharField(max_length=150, blank=True, null=True,verbose_name="Telefone")
    googleanalytics = models.CharField(max_length=150, blank=True, null=True,verbose_name="Google Analytics Stream Measurement ID")
    notificapedido = models.BooleanField(default=False, verbose_name="Receber notifição de pedido por email")

    def __str__(self):
        return f"{self.nome} - CNPJ: {self.cnpj}"


    
class Frete(models.Model):
    cidade = models.CharField(max_length=30, verbose_name="Cidade")
    bairro = models.CharField(max_length=30, verbose_name="Bairro")
    estado = BRStateField(default="RJ", verbose_name="Estado") # coming from localflavor.br
    valor = models.DecimalField(max_digits=5, decimal_places=2, verbose_name="Valor do frete")

    def __str__(self):
        return f"Destino: {self.estado}/{self.cidade}/{self.bairro} -> Valor>: R${self.valor}"

    #modifica o savar para Normalizar o nome dos bairro/cidade. Estado ja e Normalizado 
    def save(self, *args, **kwargs):
        #precisa da funcao no networkfunctions 
        self.bairro = normalizar(self.bairro)
        self.cidade = normalizar(self.cidade)
        #Do Normal Save stuff
        super().save(*args, **kwargs)


class Tag(models.Model):
    tag = models.CharField(max_length=50, unique=True, verbose_name="Tag")
    #descricao = models.CharField(max_length=50)
     
    def __str__(self):
        return self.tag
    


#Tabelas dos Dados de produtos 
class Estoque(models.Model):
    nome = models.CharField(max_length=70, verbose_name="Nome do jogo")
    nomeludopedia = models.CharField(max_length=70, blank=True, null=True, verbose_name="Nome no weblink da Ludopedia")
    descricao = models.TextField(blank=True, null=True, verbose_name="Descrição")
    valor = models.DecimalField(max_digits=5, decimal_places=2, verbose_name="Valor do aluguel")
    tags = models.ManyToManyField(to=Tag, related_name="produtos",  blank=True, null=True,verbose_name="Tags do jogo")
    quantidade = models.PositiveSmallIntegerField(default=1,verbose_name="Total na loja")
    disponivel = models.PositiveSmallIntegerField(default=1, verbose_name="Quantidade disponivel")
    dtcadastro = models.DateTimeField(editable=False, auto_now_add=True, verbose_name="Data do cadastro do cogo")
    dtatualizacao = models.DateTimeField(auto_now=True, verbose_name="Data da última alteração")    
    #Data specific for the domain of business. 
    #Right Now it is about boardgames
    tempodejogo = models.PositiveSmallIntegerField(default=1, blank=True, null=True, verbose_name="Tempo de jogo")
    minjogadores = models.PositiveSmallIntegerField(default=2, blank=True, null=True, verbose_name="Numero minímo de jogaddores")
    maxjogadores = models.PositiveSmallIntegerField(default=4, blank=True, null=True,verbose_name="Numero máximo de jogaddores")
    idade = models.CharField(max_length=25, blank=True, null=True, verbose_name="Idade recomendada")
    #idademax = models.PositiveSmallIntegerField(blank=True, null=True, verbose_name="Idade maxima recomendada")

    def __str__(self):
        return f"{self.nome} -> Total: {self.quantidade} / Disponível: {self.disponivel} / Valor: R${self.valor}"
    
    @property
    def get_first_image(self):
            vars = Variables()
            foto_jogoQS = Imagemestoque.objects.filter(produto=self.pk).first()
            if not foto_jogoQS:
                #foto_jogo = '/static/img/upload-error.png'
                foto_jogo = vars.NO_IMAGE_PNG
            else:
                foto_jogo = foto_jogoQS.foto.url
            return foto_jogo

    @property
    def get_all_images(self):
            #returns all URL for the images for the game itself in a list
            #if there is no images, it returns a link to the no image png file
            foto_jogo = []
            foto_jogoQS = Imagemestoque.objects.filter(produto=self.pk)
            if not foto_jogoQS:
                vars = Variables()
                foto_jogo.append(vars.NO_IMAGE_PNG)
            else:
                for item in foto_jogoQS:
                    foto_jogo.append(item.foto.url)
            return foto_jogo
    
    @property
    def nomepequeno(self):
        return self.nome[:30]

class Imagemestoque(models.Model):
    produto = models.ForeignKey(Estoque, on_delete=models.CASCADE)
    foto = models.ImageField(upload_to= 'produtos/', verbose_name="Path da Imagem") # this will go into the media default path
    dtcadastro = models.DateTimeField(editable=False, auto_now_add=True)
     
    def __str__(self):
        return f'{self.produto.nome} - {self.dtcadastro}'    



#Extensao via table externa dos campos dos users e receiver to make sure both tables are saved together
#Can enter empty profile alongside new user. we need to create a form that gathers allthe infomation and fill them both
#at the same time. Leaving if for when creating the interface for user (REgister form)
class Profile(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE, primary_key=True)
    #extrafields go here
    asaasid = models.CharField(max_length=70, default='0',verbose_name="Customer ID no Asaas") 
    status = models.CharField(max_length=10, choices=profile_status_choices, default='Pendente', verbose_name="Status do usuário") #list of options from assets.choices
    nome = models.CharField(max_length=70, blank=True, null=True, verbose_name="Nome")
    phone = models.CharField(max_length= 16, blank=True, null=True, verbose_name="Telefone")
    cpf = BRCPFField(blank=True, null=True ,verbose_name="CPF") #coming from localflavor.br
    endereco = models.CharField(max_length=150, blank=True, null=True, verbose_name="Logradouro")
    complemento = models.CharField(max_length=50, blank=True, null=True, verbose_name="Número e Complemento")
    cep = BRPostalCodeField(verbose_name="CEP") #coming from localflavor.br
    bairro = models.CharField(max_length=30, blank=True, null=True, verbose_name="Bairro")
    cidade = models.CharField(max_length=30, blank=True, null=True, verbose_name="Cidade")
    estado = BRStateField(default="RJ") # coming from localflavor.br
    dtnascimento = models.DateField(blank=True, null=True, verbose_name="Data de Nascimento")
    dtcadastro = models.DateTimeField(editable=False, auto_now_add=True, verbose_name="Data de Cadastro")
    dtatualizacao = models.DateTimeField(auto_now=True, verbose_name="Data da última atualização")
    foto = models.ImageField(upload_to= 'profiles/', blank=True, null=True, verbose_name="Foto do Perfil") # this will go into the media default path
    fotorg = models.ImageField(upload_to= 'profiles/', blank=True, null=True, verbose_name="Foto da Identidade") # this will go into the media default path
    fotocpf = models.ImageField(upload_to= 'profiles/', blank=True, null=True, verbose_name="Foto do CPF") # this will go into the media default path
    fotocomprovante = models.ImageField(upload_to= 'profiles/', blank=True, null=True, verbose_name="Comprovante de Residência") # this will go into the media default path
    
    #Need this to deal with a profile with no Image
    @property
    def get_perfil_image(self):    
            if not self.foto:
                vars = Variables()
                foto_perfil = vars.NO_PROFILE_FOTO
            else:
                foto_perfil = self.foto.url
            return foto_perfil

    def __str__(self):
        return f"{self.user.username} - {self.user.email}" #- {self.status}"
    
    @property
    def prettyCPF(self):
        return '{}{}{}.{}{}{}.{}{}{}-{}{}'.format(*self.cpf)
    
    @property
    def get_username(self):
        return self.user.username
    
    @property
    @admin.display(description = "Email")
    def get_email(self):
        return self.user.email
    
    """
    #modifica o savar para Normalizar o nome dos bairro/cidade. Estado ja e Normalizado 
    def save(self, *args, **kwargs):
        self.bairro = normalizar(self.bairro)
        self.cidade = normalizar(self.cidade)
        #Do Normal Save stuff
        super().save(*args, **kwargs)
    """
    
@receiver(post_save, sender=User)
def create_user_profile(sender, instance, created, **kwargs):
    if created:
        Profile.objects.create(user=instance)

@receiver(post_save, sender=User)
def save_user_profile(sender, instance, **kwargs):
    instance.profile.save()
    


#Modelos paraos pedidos. Um base, e um contendo os itens
class Pedido(models.Model):
    cliente = models.ForeignKey(User, on_delete=models.CASCADE)
    dtpedido = models.DateTimeField(editable=False, auto_now_add=True, verbose_name="Data do pedido")
    dtatualizacao = models.DateTimeField(auto_now=True, verbose_name="Ultima atualizacao")
    #Choices for pedito state
    status = models.CharField(max_length=10, choices=pedido_status_choices, default='Submetido') #list of options from assets.choices
    formapagamento = models.CharField(max_length=15, choices=billing_type_choices, default='BOLETO', verbose_name='Forma de pagamento')
    notas = models.TextField(blank=True, null=True, verbose_name="Comentários do cliente")
    #Campos de frete - Pode ser diferente do registro do cliente
    tipoentrega = models.CharField(max_length=10, choices=pedido_frete_choices, default="Pickup", verbose_name="Tipo de entrega") #list of options from assets.choices
    endereco = models.CharField(max_length=150, blank=True, null=True, verbose_name="Endereço da entrega")
    complemento = models.CharField(max_length=50, blank=True, null=True, verbose_name="Complemento")
    cep = BRPostalCodeField(blank=True, null=True, verbose_name="CEP") #coming from localflavor.br
    bairro = models.CharField(max_length=30, blank=True, null=True, verbose_name="Bairro")
    cidade = models.CharField(max_length=30, blank=True, null=True, verbose_name="Cidade")
    estado = BRStateField(default="RJ") # coming from localflavor.br
    dataenvio = models.DateField(default=timezone.now, verbose_name="Data para envio solicitada pelo cliente")
    dataretorno = models.DateField(default=add_days(7), verbose_name="Data de retorno do pedido")
    #Fields to manage available Supply and Pedido flow
    controlpago = models.BooleanField(default=False)
    controlentregue = models.BooleanField(default=False)
    controlencerrado = models.BooleanField(default=False)
    dtpago = models.DateTimeField(blank=True, null=True, verbose_name="Data do pagemento do pedido")
    dtentregue = models.DateTimeField(blank=True, null=True, verbose_name="Data da entrega do pedido")
    dtfim = models.DateTimeField(blank=True, null=True, verbose_name="Data do encerramento do pedido")
    
    def save(self, *args, **kwargs):
        #Pick up the games that are in the Order 
        # -> May want to send notifications if disponivel < 0 or > quantidade. Need to code this
        Ordem_set = Ordem.objects.values_list("produto__pk", "quantidade").filter(pedido=self.pk)        
        #Now we will make changes to the estoque fields and updade the state tracking fieldsfor each game in the order
        if (not self.controlpago and self.status in ["Pago"]):
            self.controlpago = True #mark as removed from estoque internally
            self.dtpago = timezone.now()
        if (not self.controlentregue and self.status in ["Entregue"]):
            self.controlentregue = True #mark as removed from estoque internally
            self.dtentregue= timezone.now()
            for item in Ordem_set:
                produto_item = Estoque.objects.get(pk=item[0])
                produto_item.disponivel -= item[1]
                produto_item.save()
        if (not self.controlencerrado and self.status in ["Retornado","Finalizado"]):
           self.controlencerrado = True #mark as returned to estoque internally
           self.dtfim= timezone.now()
           for item in Ordem_set:
                produto_item = Estoque.objects.get(pk=item[0])
                produto_item.disponivel += item[1]
                produto_item.save()
        #Do Normal Save stuff
        super().save(*args, **kwargs)
        
    def valorproduto(self):
        Ordem_set = Ordem.objects.values_list("produto__valor", "quantidade").filter(pedido=self.pk)        
        total = 0.0
        for item in Ordem_set:
            total += float(item[0])*int(item[1])
        return f'{total:.2f}'
    valorproduto.short_description = "Valor dos Produtos"
    
    def valorfrete(self):
        if self.tipoentrega == "Pickup":
            #Nao tem frete
            return 0.0
        #Nesse caso tera frete
        #Procura Estado, cidade e bairro na lista de fretes:
        #Precisa normalizar para evitar erros de acento
        normal_cidade = normalizar(self.cidade)
        normal_bairro = normalizar(self.bairro)
        Frete_set = Frete.objects.values_list("valor").filter(estado=self.estado, cidade=normal_cidade, bairro=normal_bairro)        
        #Testa se esta vazio ou nao a resposta, se esta volta um valor alto p indicar o erro
        if Frete_set.exists():
            #Se tiver entrada dupla, sempre manda a 1a retornada
            for item in Frete_set:
                return float(item[0])
        else:
            #This value here is to indicate that something is wrong, and we did not find a match para o frete
            return 9999.0
    valorfrete.short_description = "Valor do Frete"
    
    @property
    def valorfrete_str(self):
        return f'{float(self.valorfrete()):.2f}'
        
    def valortotal(self):
        return f'{float(self.valorproduto()) + float(self.valorfrete()):.2f}'
    valortotal.short_description = "Valor Total da Ordem"
        
    def __str__(self):
        return f"Pedido: {self.pk} - {self.cliente.username} - {self.cliente.email} - {self.status}"

class Ordem(models.Model):
    pedido = models.ForeignKey(Pedido, on_delete=models.CASCADE)
    produto = models.ForeignKey(Estoque, on_delete=models.CASCADE)
    quantidade = models.PositiveSmallIntegerField(default=1, blank=True, null=True)

    def __str__(self):
        return f"Pedido: {self.pedido.pk} - {self.produto.nome} - {self.quantidade}"



#Modelo para a sessao do FAQ
class FAQ(models.Model):
    pergunta = models.TextField(blank=True, null=True, verbose_name="Pergunta")
    resposta = models.TextField(blank=True, null=True, verbose_name="Resposta")
    
    def __str__(self):
        return f"{self.pergunta}"



#Modelo para gerenciar o cart
class Cart(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    product = models.ForeignKey(Estoque, on_delete=models.CASCADE)
    quantity = models.IntegerField(default=1)

    def __str__(self):
        return f"{self.user} - Quant: {self.quantity} - Jogo {self.product.nome} - Valor R${self.product.valor}"

    #def get_absolute_url(self):
    #    return reverse("cart:cart_detail")
    @property
    def valor_cart(self):
        return self.quantity*self.product.valor
    
    @property
    def get_first_image(self):
            vars = Variables()
            foto_jogoQS = Imagemestoque.objects.filter(produto=self.product.pk).first()
            if not foto_jogoQS:
                foto_jogo = vars.NO_IMAGE_PNG
            else:
                foto_jogo = foto_jogoQS.foto.url
            return foto_jogo

    
#Class for pagamento from Asaas
#Boleto e gerar o boleto e enviar via email
#cartao e gerar e redirecionar para entrada do cartao via o site do Asaas
#decidir como sera o callback em caso de sucesso, e em caso de falha. especialmente para pagamento em pix.
class Pagamento(models.Model):
    pedido = models.ForeignKey(Pedido, on_delete=models.CASCADE, verbose_name="Pedido") # Esse vai como externalreference no JSON e texto no description
    cliente = models.ForeignKey(User, on_delete=models.CASCADE, verbose_name="Cliente") # Para eu pegar a ID do cliente no asaas (tem um campo novo no model.Profile)
    paymentID = models.CharField(max_length=70, default='0', verbose_name="ID do pagamenta no Asaas") # string of 0 indicates that there is no payment submitted to Asaas
    status = models.CharField(max_length=45, choices=pagamento_status_choices, default='PENDING', verbose_name="Status do pagamento")
    dtsubmetido = models.DateTimeField(editable=False, auto_now_add=True, verbose_name="Data submetido")
    dtatualizacao = models.DateTimeField(auto_now=True, verbose_name="Ultima atualizacao")
    controlpago = models.BooleanField(default=False)
    dtdue = models.DateTimeField(blank=True, null=True, verbose_name="Data do vencimento do pagamento") # dueDate no Json
    dtpago = models.DateTimeField(blank=True, null=True, verbose_name="Data do pagamento")
    valorrecebido = models.DecimalField(max_digits=5, decimal_places=2, verbose_name="Valor Recebido") #Vem do retorno - netValue no Json
    valorcobranca = models.DecimalField(max_digits=5, decimal_places=2, verbose_name="Valor Cobrado") #Vem do pedido - totalValue no Json
    billingtype = models.CharField(max_length=11, choices=billing_type_choices, default='Submetido') # billingType no Json
    remoteip = models.CharField(max_length=20, default='127.0.0.1', verbose_name="Endereço IP do cliente ") # remopteIP no Json (tem de capturar o IP p cobrar via cartao de credito)
    #Coisas importantes dependendo do tipo de pagamento para uso em caso de o cliente pedir:
    transactionreceipturl = models.CharField(max_length=70, blank=True, null=True, verbose_name="Link para o pagamento no Asaas") 
    invoiceurl = models.CharField(max_length=70, blank=True, null=True, verbose_name="Link para o comprovante no Asaas") 
    retornobancario = models.TextField(blank=True, null=True, default='', verbose_name="ID dos eventos no Asaas") #campo para manter todos os eventos gerados pelo asaas e garantir nao processar o mesmo evento 2 vezes
    #boletourl = models.CharField(max_length=70, blank=True, null=True) #bankslipURL
    #pixtransaction = models.CharField(max_length=70, blank=True, null=True)
    
    @property
    @admin.display(description = "Tarifa")
    def tarifa(self):
        return self.valorcobranca - self.valorrecebido
    
    @property
    @admin.display(description = "Pedido")
    def get_pedido(self):
        return self.pedido.pk
    
    
    def __str__(self):
        return f"{self.paymentID} - {self.pedido.pk} - {self.cliente.username} - Valor R${self.valorcobranca} - Tarifa R${self.tarifa}"
                                 